import 'package:data_connection_checker/data_connection_checker.dart';

import '../../core/core.dart';
import '../../m4e_forms/m4e_forms.dart';
import '../../m4e_sdk_initialize.dart';
import '../../models/models.dart';
import '../api/apis.dart';

/// Exposes SDK Transaction Api methods
///
/// Accessing the instantiated `m4eTransaction` methods
/// is done by:
///
/// ```
/// final _m4eTransaction = M4eTransaction();
///
/// await _m4eTransaction.initiateTransaction(M4eTransactionForm m4eTransactionForm);
/// ```
///
class M4eTransactionService {
  final TransactionApi _transactionApi;
  final DataConnectionChecker _connectionChecker;

  M4eTransactionService({
    TransactionApi transactionApi,
    DataConnectionChecker connectionChecker,
  })  : _transactionApi = transactionApi ?? m4eSDK<TransactionApi>(),
        _connectionChecker =
            connectionChecker ?? m4eSDK<DataConnectionChecker>();

  /// Requires `[M4eTransactionForm]` as parameter
  ///
  /// Throws `[M4eServerException]` if call to server fails
  ///
  /// Throws `[M4eServerException]` with `NO_INTERNET_CONNECTION` error
  /// code device is not connected to the internet
  ///
  Future<M4eTransactionResult> initiateTransaction(
      M4eTransactionForm transaction) async {
    if ((await _connectionChecker.hasConnection) ?? false) {
      try {
        final _transaction =
            await _transactionApi.initiateTransaction(transaction);

        return _transaction;
      } catch (e) {
        rethrow;
      }
    }

    throw M4eExceptionMessages.kNoInternetConnectionException;
  }

  /// Returns `[M4eTransactionList]` of authenticated m4e user.
  ///
  /// Throws `[M4eServerException]` if call to server fails
  ///
  /// Throws `[M4eServerException]` with `NO_INTERNET_CONNECTION` error
  /// code device is not connected to the internet
  ///
  Future<M4eTransactionList> getUserTransactions(M4eUniqueId userId) async {
    if ((await _connectionChecker.hasConnection) ?? false) {
      try {
        final _transactionList =
            await _transactionApi.getUserTransactions(userId);

        if (_transactionList.transactions.isEmpty) {
          throw M4eExceptionMessages.kNotFoundException;
        }

        return _transactionList;
      } catch (e) {
        rethrow;
      }
    }

    throw M4eExceptionMessages.kNoInternetConnectionException;
  }

  /// Returns `[M4eTransactionList]` of authenticated m4e user.
  ///
  /// Throws `[M4eServerException]` if call to server fails
  ///
  /// Throws `[M4eServerException]` with `NO_INTERNET_CONNECTION` error
  /// code device is not connected to the internet
  ///
  Future<M4eTransactionList> getWalletTransactions(M4eUniqueId walletId) async {
    if ((await _connectionChecker.hasConnection) ?? false) {
      try {
        final _transactionList =
            await _transactionApi.getWalletTransactions(walletId);

        if (_transactionList.transactions.isEmpty) {
          throw M4eExceptionMessages.kNotFoundException;
        }

        return _transactionList;
      } catch (e) {
        rethrow;
      }
    }

    throw M4eExceptionMessages.kNoInternetConnectionException;
  }
}
