import 'package:data_connection_checker/data_connection_checker.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:m4e_sdk/src/core/core.dart';
import 'package:m4e_sdk/src/data/api/apis.dart';
import 'package:m4e_sdk/src/data/services/services.dart';
import 'package:m4e_sdk/src/m4e_forms/m4e_forms.dart';
import 'package:m4e_sdk/src/models/models.dart';
import 'package:matcher/matcher.dart';
import 'package:mockito/mockito.dart';

class MockUserApi extends Mock implements UserApi {}

class MockWalletApi extends Mock implements WalletApi {}

class MockPermissionApi extends Mock implements PermissionApi {}

class MockDataConnectionChecker extends Mock implements DataConnectionChecker {}

main() {
  M4eWalletService _m4eWalletService;

  MockUserApi _mockUserApi;
  MockWalletApi _mockWalletApi;
  MockPermissionApi _mockPermissionApi;
  MockDataConnectionChecker _mockDataConnectionChecker;

  setUp(() {
    _mockUserApi = MockUserApi();
    _mockWalletApi = MockWalletApi();
    _mockPermissionApi = MockPermissionApi();
    _mockDataConnectionChecker = MockDataConnectionChecker();

    _m4eWalletService = M4eWalletService(
      userApi: _mockUserApi,
      walletApi: _mockWalletApi,
      permissionApi: _mockPermissionApi,
      connectionChecker: _mockDataConnectionChecker,
    );
  });

  final _uniqueId =
      M4eUniqueId.fromUniqueString('46e87882-4ff8-4770-9846-870fbbcc35e0');
  final _otherUniqueId =
      M4eUniqueId.fromUniqueString('46e87882-4ff8-4770-9846-870fbbcc35e1');

  final _authToken = M4eAuthToken(id: _uniqueId, token: 'test token');

  final _phoneNumber =
      M4ePhoneNumber(country: 'GHANA', number: '+233708788110');

  final _m4eUser = M4eUser(
    id: _uniqueId,
    shortName: 'test',
    firstName: 'test',
    lastName: 'test',
    phoneNumber: _phoneNumber,
    joined: DateTime.now(),
  );

  final _m4eOtherUser = M4eUser(
    id: _otherUniqueId,
    shortName: 'test',
    firstName: 'test',
    lastName: 'test',
    phoneNumber: _phoneNumber,
    joined: DateTime.now(),
  );

  final _tWallets = <M4eWallet>[
    M4eWallet(
      id: M4eUniqueId.fromUniqueString('46e87882-4ff8-4770-9846-870fbbcc35e0'),
      walletInfo: M4eWalletInfo(label: 'W1', currency: M4eCurrencyType.ghs()),
    ),
    M4eWallet(
      id: M4eUniqueId.fromUniqueString('9e8b5ecc-9f7e-4e06-ac3d-05b11a91da88'),
      walletInfo: M4eWalletInfo(label: 'W2', currency: M4eCurrencyType.ghs()),
    ),
  ];

  final _tWalletList = M4eWalletList(wallets: _tWallets);

  final _tM4eWalletForm =
      M4eWalletForm(label: 'test', currency: M4eCurrencyType.ghs());

  final _ownerWalletAccessPermissionsSet = <M4eAccessPermissionSet>[
    EntityWalletAccess.view,
    EntityWalletAccess.grant,
    EntityWalletAccess.alter,
  ];

  final _otherUsersWalletAccessPermissionsSet = <M4eAccessPermissionSet>[
    EntityWalletAccess.see,
    EntityWalletAccess.pay,
  ];

  // Single search clause ======================================================================================
  final _singleSearchFormTypeCollection = SingleClauseSearchForm(
    clause: SearchClause(
      subjectId: _uniqueId,
      type: M4eAccessPermissionType.ENTITY_WALLET,
      encoding: _ownerWalletAccessPermissionsSet,
    ),
  );

  // Multiple search clauses ====================================================================================
  final _ownerWalletsAccessSearchClause = SearchClause(
    subjectId: _otherUniqueId,
    type: M4eAccessPermissionType.ENTITY_WALLET,
    encoding: _ownerWalletAccessPermissionsSet,
  );

  // accessing wallets current user has [SEE, PAY] ENTITY_WALLET permissions on
  final _currentUsersWalletAccessSearchClause = SearchClause(
    subjectId: _uniqueId,
    type: M4eAccessPermissionType.ENTITY_WALLET,
    encoding: _otherUsersWalletAccessPermissionsSet,
  );

  final _multipleSearchFormTypeCollection = MultipleClauseSearchForm(
    conjunction: M4eSearchConjunctionType.AND,
    clauses: <SearchClause>[
      _ownerWalletsAccessSearchClause,
      _currentUsersWalletAccessSearchClause,
    ],
  );

  group('m4e-wallet-test | ', () {
    group('getWallets | ', () {
      test('should verify device connectivity status', () async {
        // arrange
        when(_mockDataConnectionChecker.hasConnection)
            .thenAnswer((_) async => true);
        when(_mockWalletApi.queryUserWallets(any))
            .thenAnswer((_) async => _uniqueId);
        when(_mockWalletApi.runSearchQuery(any))
            .thenAnswer((_) async => _tWalletList);

        // act
        await _m4eWalletService.getWallets(currentUserId: _uniqueId);

        // assert
        verify(_mockDataConnectionChecker.hasConnection);
      });

      group('isConnected', () {
        setUp(() {
          when(_mockDataConnectionChecker.hasConnection)
              .thenAnswer((_) async => true);
          when(_mockWalletApi.queryUserWallets(any))
              .thenAnswer((_) async => _uniqueId);
          when(_mockWalletApi.runSearchQuery(any))
              .thenAnswer((_) async => _tWalletList);
        });

        group('current-user-wallet', () {
          setUp(() {
            when(_mockUserApi.getM4eUser(any))
                .thenAnswer((_) async => _m4eUser);
          });

          test(
              'should call [M4eUserApi] with [M4ePhoneNumber] as parameter to retrive [M4eUser]',
              () async {
            // act
            await _m4eWalletService.getWallets(
                currentUserId: _uniqueId, otherUserPhoneNumber: _phoneNumber);

            // assert
            verify(_mockUserApi.getM4eUser(_phoneNumber));
          });

          test('should return [M4eWalletList] on success call', () async {
            // arrange
            when(_mockWalletApi.queryUserWallets(any))
                .thenAnswer((_) async => _uniqueId);

            // act
            final _result =
                await _m4eWalletService.getWallets(currentUserId: _uniqueId);

            // assert
            expect(_result, equals(_tWalletList));
          });

          test(
              'should return [WalletApi.queryUserWallets] with appropriate parameters (query with single-search-clause when [currentUserId] is passed in)',
              () async {
            // arrange
            when(_mockWalletApi.queryUserWallets(any))
                .thenAnswer((_) async => _uniqueId);

            // act
            final _result =
                await _m4eWalletService.getWallets(currentUserId: _uniqueId);

            // assert
            expect(_result, equals(_tWalletList));
            verify(_mockWalletApi
                .queryUserWallets(_singleSearchFormTypeCollection));
          });

          test(
              'should return [WalletApi.queryUserWallets] with appropriate parameters (query with single-search-clause when passed in [currentUserId] equals [otherUserId])',
              () async {
            // arrange
            when(_mockWalletApi.queryUserWallets(any))
                .thenAnswer((_) async => _uniqueId);

            // act
            final _result = await _m4eWalletService.getWallets(
                currentUserId: _uniqueId, otherUserPhoneNumber: _phoneNumber);

            // assert
            expect(_result, equals(_tWalletList));
            verify(_mockWalletApi
                .queryUserWallets(_singleSearchFormTypeCollection));
          });

          test('should call [WalletApi.runSearchQuery] after creating query',
              () async {
            // arrange
            when(_mockWalletApi.queryUserWallets(any))
                .thenAnswer((_) async => _uniqueId);

            // act

            await _m4eWalletService.getWallets(currentUserId: _uniqueId);

            // assert
            verify(_mockWalletApi.runSearchQuery(_uniqueId));
          });

          test('should re-throw a ServerException when server call fails',
              () async {
            // arrange
            when(_mockWalletApi.queryUserWallets(any))
                .thenThrow(M4eExceptionMessages.kUndefinedException);

            // act
            final _call = _m4eWalletService.getWallets;

            // assert
            expect(() => _call(currentUserId: _uniqueId),
                throwsA(TypeMatcher<M4eServerException>()));
          });

          test(
              'should re-throw a ServerException when server call to [WalletApi.runSearchQuery] fails',
              () async {
            // arrange
            when(_mockWalletApi.queryUserWallets(any))
                .thenAnswer((_) async => _uniqueId);
            when(_mockWalletApi.runSearchQuery(any))
                .thenThrow(M4eExceptionMessages.kUndefinedException);

            // act
            final _call = _m4eWalletService.getWallets;

            // assert
            expect(() => _call(currentUserId: _uniqueId),
                throwsA(TypeMatcher<M4eServerException>()));
          });
        });

        group('other user wallets', () {
          setUp(() {
            when(_mockUserApi.getM4eUser(any))
                .thenAnswer((_) async => _m4eOtherUser);
          });

          test(
              'should call [WalletApi.queryUserWallets] with appropriate parameters (query with multiple-search-clause when [currentUserId, otherUserId] is passed in)',
              () async {
            // arrange
            when(_mockWalletApi.queryUserWallets(any))
                .thenAnswer((_) async => _uniqueId);

            // act
            final _result = await _m4eWalletService.getWallets(
                currentUserId: _uniqueId, otherUserPhoneNumber: _phoneNumber);

            // assert
            expect(_result, equals(_tWalletList));
            verify(_mockWalletApi
                .queryUserWallets(_multipleSearchFormTypeCollection));
          });
        });
      });

      group('isDisconnected', () {
        setUp(() {
          when(_mockDataConnectionChecker.hasConnection)
              .thenAnswer((_) async => false);
        });
        test(
            'should throw [ServerException] with [NO_INTERNET_CONNECTION] error code',
            () async {
          // arrange
          when(_mockWalletApi.queryUserWallets(any))
              .thenThrow(M4eExceptionMessages.kNoInternetConnectionException);

          // act
          final _call = _m4eWalletService.getWallets;

          // assert
          expect(() => _call(currentUserId: _uniqueId),
              throwsA(TypeMatcher<M4eServerException>()));
        });
      });
    });

    group('createWallet | ', () {
      test('should verify device connectivity status', () async {
        // arrange
        when(_mockDataConnectionChecker.hasConnection)
            .thenAnswer((_) async => true);
        // act
        await _m4eWalletService.createWallet(_uniqueId, _tM4eWalletForm);

        // assert
        verify(_mockDataConnectionChecker.hasConnection);
      });

      group('isConnected', () {
        setUp(() {
          when(_mockDataConnectionChecker.hasConnection)
              .thenAnswer((_) async => true);
        });

        test('should call [M4eWalletApi] to createWallet with right parameters',
            () async {
          // assert
          when(_mockWalletApi.createWallet(any, any))
              .thenAnswer((_) async => _uniqueId);

          // act
          final _result =
              await _m4eWalletService.createWallet(_uniqueId, _tM4eWalletForm);

          // assert
          verify(_mockWalletApi.createWallet(_authToken.id, _tM4eWalletForm));
          expect(_result, equals(_uniqueId));
        });

        test('should re-throw a ServerException when server call fails',
            () async {
          // arrange
          when(_mockWalletApi.createWallet(any, any))
              .thenThrow(M4eExceptionMessages.kUndefinedException);

          // act
          final _call = _m4eWalletService.createWallet;

          // assert
          expect(() => _call(_uniqueId, _tM4eWalletForm),
              throwsA(TypeMatcher<M4eServerException>()));
        });
      });

      group('isDisconnected', () {
        setUp(() {
          when(_mockDataConnectionChecker.hasConnection)
              .thenAnswer((_) async => false);
        });
        test(
            'should throw [ServerException] with [NO_INTERNET_CONNECTION] error code',
            () async {
          // arrange
          when(_mockWalletApi.createWallet(any, any))
              .thenThrow(M4eExceptionMessages.kNoInternetConnectionException);

          // act
          final _call = _m4eWalletService.createWallet;

          // assert
          expect(() => _call(_uniqueId, _tM4eWalletForm),
              throwsA(TypeMatcher<M4eServerException>()));
        });
      });

      test('should call wallet-api with approprite paraameters', () {
        final _permissions = <M4eAccessPermissionSet>[
          EntityWalletAccess.see,
          EntityWalletAccess.view,
          EntityWalletAccess.pay,
          EntityWalletAccess.grant,
          EntityWalletAccess.alter,
          EntityWalletAccess.create,
          EntityWalletAccess.destroy,
        ];

        String _permissionUrl =
            '/wallets/access?entity=${_uniqueId.toString()}';

        for (final item in _permissions) {
          _permissionUrl += '&access=${item.toString()}';
        }

        print(_permissionUrl);
      });
    });
  });
}
